#!/usr/bin/env python
# -*- coding: utf-8 -*-
'''This module provides functions for assembly of sequences by homologous recombination and other
related techniques. Given a list of sequences (Dseqrecords), all sequences will be analyzed for
overlapping regions of DNA (common substrings).

The assembly algorithm is based on graph theory where each overlapping region forms a node and 
sequences separating the overlapping regions form edges.

'''

import itertools
import networkx as nx
import operator
import random

from copy import copy
from textwrap import dedent
from collections import defaultdict
from collections import namedtuple

from Bio.SeqFeature import ExactPosition
from Bio.SeqFeature import FeatureLocation
from Bio.SeqFeature import SeqFeature

from pydna.dsdna import Dseq
from pydna.dsdna import Dseqrecord
from pydna._simple_paths8 import all_simple_paths_edges, all_circular_paths_edges
from findsubstrings_suffix_arrays_python import common_sub_strings
from findsubstrings_suffix_arrays_python import terminal_overlap

from orderedset import OrderedSet


class Fragment(Dseqrecord):
    '''This class holds information about a DNA fragment in an assembly.
    This class is instantiated by the :class:`Assembly` class and is not 
    meant to be instantiated directly.
    
    '''
    
    def __init__(self, record, start1    = 0,
                               end1      = 0,
                               start2    = 0,
                               end2      = 0,
                               alignment = 0,
                               i         = 0, *args, **kwargs):
        
        super(Fragment, self).__init__(record, *args, **kwargs)
        
        self.start1             = start1
        self.end1               = end1
        self.left_overlap_size  = end1-start1
        self.start2             = start2
        self.end2               = end2
        self.right_overlap_size = end2-start2
        self.alignment          = alignment
        self.i                  = i

    def __str__(self):
        return ("Fragment alignment {}\n").format(self.alignment)+super(Fragment, self).__str__()

class Contig(Dseqrecord):
    '''This class holds information about a DNA assembly. This class is instantiated by 
    the :class:`Assembly` class and is not meant to be instantiated directly.
    
    '''
    
    def __init__(self, 
                 record, 
                 source_fragments=[], 
                 *args, **kwargs):
                       
        super(Contig, self).__init__(record, *args, **kwargs)                
        self.source_fragments = source_fragments
        self.number_of_fragments = len(self.source_fragments)
    
    def detailed_figure(self):
        '''Synonym of :func:`detailed_fig`'''
        return self.detailed_fig()
        
    def detailed_fig(self):

        fig=""       
        for s in self.source_fragments:
            fig +="{}{}\n".format(" "*s.alignment, str(s.seq))        
        return fig
        
    def small_figure(self):
        '''Synonym of :func:`small_fig`'''
        return self.small_fig()
    
    def small_fig(self):
        ''' 
        Returns a small ascii representation of the assembled fragments. Each fragment is 
        represented by:
                
        ::
        
         Size of common 5' substring|Name and size of DNA fragment| Size of common 5' substring
        
        Linear:
        
        ::
        
          frag20| 6
                 \\/
                 /\\
                  6|frag23| 6
                           \\/
                           /\\
                            6|frag14
        
        
        Circular:
        
        ::
        
          -|2577|61
         |       \\/
         |       /\\
         |       61|5681|98
         |               \\/
         |               /\\
         |               98|2389|557
         |                       \\/
         |                       /\\
         |                       557-
         |                          |
          --------------------------
        
        
        '''

        if self.linear:
            '''
            frag20| 6
                   \/
                   /\
                    6|frag23| 6
                             \/
                             /\
                              6|frag14
            '''
            f = self.source_fragments[0]
            space2 = len(f.name)
            
            
            fig = ("{name}|{o2:>2}\n"
                   "{space2} \/\n"
                   "{space2} /\\\n").format(name = f.name,
                                            o2 = f.right_overlap_size,
                                            space2 = " "*space2)
            space = len(f.name)

            for f in self.source_fragments[1:-1]:        
                name= "{o1:>2}|{name}|".format(o1   = f.left_overlap_size,
                                               name = f.name)
                space2 = len(name)
                fig +=("{space} {name}{o2:>2}\n"
                       "{space} {space2}\/\n"
                       "{space} {space2}/\\\n").format( name = name,
                                                        o2 = f.right_overlap_size,
                                                        space = " "*space,
                                                        space2 = " "*space2)
                space +=space2 
            f = self.source_fragments[-1]
            fig += ("{space} {o1:>2}|{name}").format(name = f.name,
                                                    o1 = f.left_overlap_size,
                                                    space = " "*(space))
                
                
             
        else:
            '''
             -|2577|61
            |       \/
            |       /\
            |       61|5681|98
            |               \/
            |               /\
            |               98|2389|557
            |                       \/
            |                       /\
            |                       557-
            |                          |
             --------------------------
            '''
            f = self.source_fragments[0]        
            space = len(f.name)+3        
            fig =(" -|{name}|{o2:>2}\n"
                  "|{space}\/\n"
                  "|{space}/\\\n").format(name = f.name,
                                           o2 = f.right_overlap_size,
                                           space = " "*space)             
            for f in self.source_fragments[1:]:               
                name= "{o1:>2}|{name}|".format(o1 = f.left_overlap_size,
                                                      name = f.name)
                space2 = len(name)
                fig +=("|{space}{name}{o2:>2}\n"
                       "|{space}{space2}\/\n"
                       "|{space}{space2}/\\\n").format(o2 = f.right_overlap_size,
                                                       name = name,
                                                       space = " "*space,
                                                       space2 = " "*space2)
                space +=space2 
                                           
            fig +="|{space}{o1:>2}-\n".format(space=" "*(space), o1=self.source_fragments[0].left_overlap_size)    
            fig +="|{space}   |\n".format(space=" "*(space))    
            fig +=" {space}".format(space="-"*(space+3))             
        return dedent(fig)

class Assembly(object):
    '''Assembly of a list of linear DNA fragments into linear or circular constructs.   
    Accepts a list of Dseqrecords (source fragments) to initiate an Assembly object. 
    Several methods are available for analysis of overlapping sequences, graph construction
    and assembly.

    Parameters
    ----------

    dsrecs : list
        a list of Dseqrecord objects.
        
    Examples
    --------
    
    >>> from pydna import Assembly, Dseqrecord
    >>> a = Dseqrecord("acgatgctatactgCCCCCtgtgctgtgctcta")
    >>> b = Dseqrecord("tgtgctgtgctctaTTTTTtattctggctgtatc")
    >>> c = Dseqrecord("tattctggctgtatcGGGGGtacgatgctatactg")
    >>> x = Assembly((a,b,c), limit=14)
    >>> x.analyze_overlaps()
    '6 sequences analyzed of which 3 have shared homologies with totally 3 overlaps'
    >>> x.create_graph()
    "A graph with 5', 3' and 3 internal nodes was created"
    >>> x.assemble_circular_from_graph()
    '1 circular products were formed'
    >>> x
    Assembly object:
    Sequences........................: 33, 34, 35
    Sequences with shared homologies.: 33, 34, 35
    Homology limit (bp)..............: 14
    Number of overlaps...............: 3
    Nodes in graph...................: 5
    Assembly protocol................: overlaps
    Circular products................: 59
    Linear products..................: No linear products
    >>> x.circular_products
    [Dseqrecord(o59)]
    >>> x.circular_products[0].seq.watson
    'CCCCCtgtgctgtgctctaTTTTTtattctggctgtatcGGGGGtacgatgctatactg'


    '''

    def __init__(self, dsrecs, limit = 25):
        self.dsrecs    = dsrecs
        ''' Sequences fed to this class is stored in this property'''
        self.max_nodes = len(self.dsrecs)
        ''' The max number of nodes allowed. This can be reset to some other value'''
        self.limit     = limit
        ''' The shortest common sub strings to be considered '''
        
        for dr in self.dsrecs:
            if dr.name in ("",".", "<unknown name>", None):
                dr.name = "frag{}".format(len(dr))

    def analyze_overlaps(self, limit=None):
        '''This method carries out a pairwise analysis for common sub strings (shared homologous sequences) 
        among DNA fragments stored in the dsrecs property. 
        
        * Common sub strings need to be equal to or longer than limit. 
        * Common sub strings can be present anywhere within the sequences stored in the dsrecs property. 
        
        The analyzed_dsrec property will contain copies of the sequences in the dsrecs property that have 
        common substrings. Common sub strings are added as features to the sequences in the analyzed_dsrec 
        property.
        
        
        
        Parameters
        ----------

        limit : int
            The shortest shared sub sequence to be considered. This will also set the limit property.
        
        See also
        --------
        pydna.assembly.Assembly.analyze_terminal_overlaps
            
        '''
        
        if limit:
            self.limit = limit
        self.protocol = "overlaps"
        return self._analyze_overlaps(common_sub_strings)

    def analyze_terminal_overlaps(self, limit=None):
        '''This method carries out a pairwise analysis for shared sub sequences equal to or longer than
        limit. The difference between this method and :func:`analyze_overlaps` is that shared sequences 
        must start or end at the extremities of the source fragments.
        
        Parameters
        ----------

        limit : int
            The shortest shared sub sequence to be considered. This will also set the limit property.
            
        See also
        --------
        pydna.assembly.Assembly.analyze_overlaps
             
        '''
        if limit:
            self.limit = limit
        self.protocol = "terminal overlaps"
        return self._analyze_overlaps(terminal_overlap)
        
    def create_graph(self):
        '''Creates a graph from analyzed source fragments. Nodes in the graph are shared sub sequences
        found as features in the source sequences. Edges are the sequences that connect the shared 
        sub sequences in each source fragment. Two additional nodes called 5' and 3' are also added to 
        represent DNA ends.        
        '''     
        return self._create_graph()

    def assemble_circular_from_graph(self):
        '''Assembles all circular products from the graph created by :func:`create_graph`'''
        self.circular_products = self._circ()
        return "{} circular products were formed".format(len(self.circular_products))

    def assemble_linear_from_graph(self):
        '''Assembles all linear products from the graph created by :func:`create_graph`'''
        self.linear_products = self._lin()
        return "{} linear products were formed".format(len(self.linear_products))
        
    def assemble_hr_circular(self, limit=None):
        '''Convenience method to assembles all possible circular homologous recombination (hr) products
        expected from in-vivo homologous recombination by the following steps:
        
        1. Analysis of source fragments using the :func:`analyze_overlaps` method        
        2. Construction of a graph using the :func:`create_graph` method        
        3. Assembly of possible circular products using the :func:`assemble_circular_from_graph` method        
        
        The property "circular_products" is initiated with a list of :class:`Fragment` objects.
               
        '''
        if limit:
            self.limit = limit            
        self.analyze_overlaps()
        self.create_graph()
        self.assemble_circular_from_graph()
        self.protocol = "homologous recombination, circular"
        return "{} circular products were formed".format(len(self.circular_products))
        
    def assemble_hr_linear(self, limit = None):
        '''Convenience method to assembles all possible linear homologous recombination (hr) products
        expected from in-vivo homologous recombination by the following steps:
        
        1. Analysis of source fragments using the :func:`analyze_overlaps` method        
        2. Construction of a graph using the :func:`create_graph` method        
        3. Assembly of possible linear products using the :func:`assemble_linear_from_graph` method        
        
        The property "linear_products" is initiated with a list of :class:`Fragment` objects.
       
        '''
        if limit:
            self.limit = limit            
        self.analyze_overlaps()
        self._create_graph()
        self.linear_products = self._lin()
        self.protocol = "homologous recombination, linear"
        return "{} linear products were formed".format(len(self.linear_products))
        
    def assemble_gibson_circular(self, limit = None):
        '''Convenience method to assembles all possible circular Gibson assembly products
        expected from in-vivo homologous recombination by the following steps:
        
        1. Analysis of source fragments using the :func:`analyze_terminal_overlaps` method        
        2. Construction of a graph using the :func:`create_graph` method        
        3. Assembly of possible circular products using the :func:`assemble_circular_from_graph` method        
        
        The property "circular_products" is initiated with a list of :class:`Fragment` objects.
               
        '''
        if limit:
            self.limit = limit            
        self.analyze_terminal_overlaps()
        self._create_graph()
        self.circular_products = self._circ()
        self.protocol = "gibson assembly, circular"
        return "{} circular products were formed with the following sizes (bp): {}".format(len(self.circular_products), 
                                                                                           ", ".join(str(len(x)) for x in self.circular_products))  

    def assemble_gibson_linear(self, limit = None):
        '''Convenience method to assembles all possible linear Gibson assembly products
        expected from in-vivo homologous recombination by the following steps:
        
        1. Analysis of source fragments using the :func:`analyze_terminal_overlaps` method        
        2. Construction of a graph using the :func:`create_graph` method        
        3. Assembly of possible circular products using the :func:`assemble_linear_from_graph` method        
        
        The property "linear_products" is initiated with a list of :class:`Fragment` objects.
               
        '''
        if limit:
            self.limit = limit            
        self.analyze_terminal_overlaps()
        self._create_graph()
        self.linear_products = self._lin()
        self.protocol = "gibson assembly, linear"
        return "{} linear products were formed with the following sizes (bp): {}".format( len(self.linear_products), 
                                                                                          ", ".join(str(len(x)) for x in self.linear_products))
    def assemble_fusion_pcr_linear(self, limit = None):
        '''Convenience method to assembles all possible linear fusion PCR products.
        This method is a synonym of :func:`assemble_gibson_linear` 
        '''
        self.assemble_gibson_linear(limit)
        self.protocol = "fusion PCR, linear"
        return "{} linear products were formed".format(len(self.linear_products))
        
    def assemble_slic(self, limit = None):
        '''Not implemented''' 
        self.protocol = "slic, linear"
        raise NotImplementedError
         
    def assemble_golden_gate(self, limit = None):
        '''Not implemented''' 
        self.protocol = "golden_gate"
        raise NotImplementedError

    def _create_graph(self):
        
        self.G=nx.MultiDiGraph(multiedges=True, name ="original graph" , selfloops=False)
        self.G.add_node( '5' )
        self.G.add_node( '3' )
       
        for i, dsrec in enumerate(self.analyzed_dsrecs):

            overlaps = sorted( {f.qualifiers['chksum'][0]:f for f in dsrec.features 
                                if f.type=='overlap'}.values(),
                               key = operator.attrgetter('location.start'))        

            if overlaps:
                overlaps = ([SeqFeature(FeatureLocation(0, 0),
                             type = 'overlap',
                             qualifiers = {'chksum':['5']})]+
                             overlaps+
                            [SeqFeature(FeatureLocation(len(dsrec),len(dsrec)),
                                        type = 'overlap',
                                        qualifiers = {'chksum':['3']})])
                
                for olp1, olp2 in itertools.combinations(overlaps, 2):

                    n1 = olp1.qualifiers['chksum'][0]
                    n2 = olp2.qualifiers['chksum'][0]
                    
                    if n1 == '5' and n2=='3':
                        continue   
                    
                    s1,e1,s2,e2 = (olp1.location.start.position,
                                   olp1.location.end.position,
                                   olp2.location.start.position,
                                   olp2.location.end.position,)
                    
                    source_fragment = Fragment(dsrec,s1,e1,s2,e2,i)

                    self.G.add_edge( n1, n2, 
                                     frag=source_fragment,
                                     weight = s1-e1,
                                     i = i)
        return '''A graph with 5', 3' and {nodes} internal nodes was created'''.format(nodes = self.G.order()-2)


    def _analyze_overlaps(self, algorithm):
        cols = {}
        for dsrec in self.dsrecs:
            dsrec.features = [f for f in dsrec.features if f.type!="overlap"]
            dsrec.seq = Dseq(dsrec.seq.todata)
        rcs = {dsrec:dsrec.rc() for dsrec in self.dsrecs}
        matches=[]
        dsset=OrderedSet()

        for a, b in itertools.combinations(self.dsrecs, 2):
            #print len(a),len(b)
            #a,b = b,a
            match = algorithm( str(a.seq).upper(),
                               str(b.seq).upper(),
                               self.limit)
            if match:
                matches.append((a, b, match))
                dsset.add(a)
                dsset.add(b)                
            match = algorithm( str(a.seq).upper(),
                               str(rcs[b].seq).upper(),
                               self.limit)
            if match:
                matches.append((a, rcs[b], match))
                dsset.add(a)
                dsset.add(rcs[b])
                matches.append((rcs[a], b, [(len(a)-sa-le,len(b)-sb-le,le) for sa,sb,le in match]))
                dsset.add(b)
                dsset.add(rcs[a])
        
        self.no_of_olaps=0            
        
        for a, b, match in matches:
            for start_in_a, start_in_b, length in match:
                self.no_of_olaps+=1
                chksum = a[start_in_a:start_in_a+length].seguid()
                assert chksum == b[start_in_b:start_in_b+length].seguid()
                
                try:
                    fcol, revcol = cols[chksum]
                except KeyError:
                    fcol = '#%02X%02X%02X' % (random.randint(175,255),random.randint(175,255),random.randint(175,255))
                    rcol = '#%02X%02X%02X' % (random.randint(175,255),random.randint(175,255),random.randint(175,255))
                    cols[chksum] = fcol,rcol
                
                qual      = {"note"             : ["olp_{}".format(chksum)],
                             "chksum"           : [chksum],
                             "ApEinfo_fwdcolor" : [fcol],
                             "ApEinfo_revcolor" : [rcol]}
                
                if not chksum in [f.qualifiers["chksum"][0] for f in a.features if f.type == "overlap"]:            
                    a.features.append( SeqFeature( FeatureLocation(start_in_a,
                                                                   start_in_a + length),
                                                                   type = "overlap",
                                                                   qualifiers = qual))
                if not chksum in [f.qualifiers["chksum"][0] for f in b.features if f.type == "overlap"]:
                    b.features.append( SeqFeature( FeatureLocation(start_in_b,
                                                                   start_in_b + length),
                                                                   type = "overlap",
                                                                   qualifiers = qual))
        for ds in dsset:
            ds.features = sorted([f for f in ds.features], key = operator.attrgetter("location.start"))
            
        self.analyzed_dsrecs = list(dsset)

        return ( "{number_of_seqs} sequences analyzed "
                 "of which {analyzed_dsrecs} have shared homologies "
                 "with totally {no_of_olaps} overlaps").format(number_of_seqs  = len(self.dsrecs)*2,
                                                               analyzed_dsrecs = len(self.analyzed_dsrecs),
                                                               no_of_olaps     = self.no_of_olaps)
    

    def _circ(self):        
        self.cG = self.G.copy()    
        self.cG.remove_nodes_from(('5','3'))
        circular_products=defaultdict(list)
        
        for pth in all_circular_paths_edges(self.cG):

            ns = min(enumerate(pth), key = lambda x:x[1][2]['i'])[0]
            
            path = pth[ns:]+pth[:ns]
            
            pred_frag = copy(path[0][2]['frag'])

            source_fragments = [pred_frag, ]
                 
            if pred_frag.start2<pred_frag.end1:
                result=pred_frag[pred_frag.start2+(pred_frag.end1-pred_frag.start2):pred_frag.end2]
            else:
                result=pred_frag[pred_frag.end1:pred_frag.end2]
            
            result.seq = Dseq(str(result.seq))
            
            for first_node, second_node, edgedict in path[1:]:
                
                f  = copy(edgedict['frag'])
                
                f.alignment =  pred_frag.alignment + pred_frag.start2- f.start1
                source_fragments.append(f)
                
                if f.start2>f.end1:
                    nxt = f[f.end1:f.end2]
                else:
                    nxt =f[f.start2+(f.end1-f.start2):f.end2]
                nxt.seq = Dseq(str(nxt.seq))
                result+=nxt
                
                pred_frag = f
                
            add=True            
            for cp in circular_products[len(result)]:
                if (str(result.seq).lower() in str(cp.seq).lower()*2
                    or
                    str(result.seq).lower() == str(cp.seq.reverse_complement()).lower()*2):
                    pass
                    add=False                
            if add:
                circular_products[len(result)].append( Contig( Dseqrecord(result, circular=True), source_fragments))

        return list(itertools.chain.from_iterable(circular_products[size] for size in sorted(circular_products, reverse=True)))
        
    def _lin(self):
        
        linear_products=defaultdict(list)

        for path in all_simple_paths_edges(self.G, '5', '3', data=True, cutoff=self.max_nodes):
            
            pred_frag = copy(path[0][2].values().pop()['frag'])
            source_fragments = [pred_frag, ]
            
            if pred_frag.start2<pred_frag.end1:
                result=pred_frag[pred_frag.start2+(pred_frag.end1-pred_frag.start2):pred_frag.end2]
            else:
                result=pred_frag[pred_frag.end1:pred_frag.end2]
            
            for first_node, second_node, edgedict in path[1:]:

                edgedict = edgedict.values().pop()
                
                f  = copy(edgedict['frag'])
                
                f.alignment =  pred_frag.alignment + pred_frag.start2- f.start1
                source_fragments.append(f)
                
                if f.start2>f.end1:
                    result+=f[f.end1:f.end2]                    
                else:
                    result+=f[f.start2+(f.end1-f.start2):f.end2]
                
                pred_frag = f

            add=True
            for lp in linear_products[len(result)]:
                if (str(result.seq).lower() == str(lp.seq).lower()
                    or
                    str(result.seq).lower() == str(lp.seq.reverse_complement()).lower()):
                    add=False
            for dsrec in self.dsrecs:
                if (str(result.seq).lower() == str(dsrec.seq).lower()
                    or
                    str(result.seq).lower() == str(dsrec.seq.reverse_complement()).lower()):
                    add=False
            if add:
                linear_products[len(result)].append(Contig( result, source_fragments))
    
        return list(itertools.chain.from_iterable(linear_products[size] for size in sorted(linear_products, reverse=True)))

    def __repr__(self):
            
        try:    
            number_of_seqs  =  ", ".join(str(len(x)) for x in self.dsrecs)
        except AttributeError:
            number_of_seqs  =  "No sequences"
        try:    
            analyzed_dsrecs = ", ".join(str(len(x)) for x in self.analyzed_dsrecs)
        except AttributeError:
            analyzed_dsrecs = "No analyzed sequences"
        try:    
            limit = self.limit        
        except AttributeError:
            limit = "No limit set"
        try:    
            no_of_olaps = self.no_of_olaps        
        except AttributeError:
            no_of_olaps = "No overlaps"
        try:
            nodes = self.G.order()
        except AttributeError:
            nodes = "No graph"  
        try:    
            pr = self.protocol 
        except AttributeError:
            pr= "No protocol"
        try:    
            cp = ", ".join(str(len(x)) for x in self.circular_products)
        except AttributeError:
            cp= "No circular products" 
        try:    
            lp = ", ".join(str(len(x)) for x in self.linear_products)
        except AttributeError:
            lp= "No linear products"       

        return   ( "Assembly object:\n"
                   "Sequences........................: {number_of_seqs}\n"
                   "Sequences with shared homologies.: {analyzed_dsrecs}\n"
                   "Homology limit (bp)..............: {limit}\n"
                   "Number of overlaps...............: {no_of_olaps}\n"
                   "Nodes in graph...................: {nodes}\n"
                   "Assembly protocol................: {pr}\n"
                   "Circular products................: {cp}\n"
                   "Linear products..................: {lp}"    ).format(number_of_seqs  = number_of_seqs,
                                                                         analyzed_dsrecs = analyzed_dsrecs,
                                                                         limit           = limit,
                                                                         no_of_olaps     = no_of_olaps,
                                                                         nodes           = nodes,
                                                                         pr              = pr,
                                                                         cp              = cp,
                                                                         lp              = lp)

if __name__=="__main__":
    import doctest
    doctest.testmod()
    
    from pydna_helper import ape
    from pydna import Dseqrecord, Dseq
    from pydna import parse
    from pydna import eq
    from pprint import pprint
    
    a = Dseqrecord(Dseq("acgatgctatactgCCCCCtgtgctgtgctcta",
                        "tgctacgatatgacGGGGGacacgacacgagat"[::-1]))
       
    b = Dseqrecord(Dseq("tgtgctgtgctctaTTTTTtattctggctgtatc",
                        "acacgacacgagatAAAAAataagaccgacatag"[::-1]))
    
    c = Dseqrecord(Dseq("tattctggctgtatcGGGGGtacgatgctatactg",
                        "ataagaccgacatagCCCCCatgctacgatatgac"[::-1]))
    
    pw = [a,c,b]
    
    a = Assembly(pw, limit=14)

    print a.assemble_hr_circular()
    
    print a.circular_products
    
    #print a.circular_products[0].small_fig()
    
    #print a.circular_products[0].seq
    
    #print '''CCCCCtgtgctgtgctctaTTTTTtattctggctgtatcGGGGGtacgatgctatactg'''
    
    #for x in a.circular_products:
        #print x.seq
    #    print x.source_fragments
    #    print 
    
    
  
    



#!/usr/bin/env python
# -*- coding: utf-8 -*-
'''Provides a class for downloading sequences from genbank.
'''
import re
import os
import urllib2
import warnings

from urlparse               import urlparse
from urlparse               import urlunparse
from Bio                    import SeqIO
from Bio                    import Entrez
from Bio.Alphabet.IUPAC     import IUPACAmbiguousDNA
from Bio.GenBank            import RecordParser
from Bio.SeqUtils.CheckSum  import seguid

from pydna.dsdna import read

class Genbank():
    '''Class to facilitate download from genbank.

    Parameters
    ----------
    users_email : string
        Has to be a valid email address. You should always tell
        Genbanks who you are, so that they can contact you.
    proxy : string, optional
        String containing a proxy url:
        "proxy = "http://umiho.proxy.com:3128"
    tool : string, optional
        Default is "pydna". This is to tell Genbank which tool you are
        using.

    Examples
    --------
    
    >>> import pydna 
    >>> gb=pydna.Genbank("me@mail.se", proxy = "http://proxy.com:3128")                  #doctest: +SKIP
    >>> rec = gb.nucleotide("L09137") # <- pUC19 from genbank                            #doctest: +SKIP
    >>> print len(rec)                                                                   #doctest: +SKIP
    2686                                                                                 #doctest: +SKIP
    '''

    def __init__(self, users_email, proxy = None, tool="pydna"):
        
        if not re.match("[A-Z0-9._%+-]+@[A-Z0-9.-]+\.[A-Z]{2,4}",users_email,re.IGNORECASE):
            raise(ValueError("Not a valid email address!"))
            
        self.email=users_email #Always tell NCBI who you are

        if proxy:
            parsed = urlparse(proxy)
            scheme = parsed.scheme
            hostname = parsed.hostname
            test = urlunparse((scheme, hostname,'','','','',))
            try:
                response=urllib2.urlopen(test, timeout=1)
            except urllib2.URLError as err:
                warnings.warn("could not contact proxy server")
            print { scheme : parsed.geturl() }
            self.proxy = urllib2.ProxyHandler({ scheme : parsed.geturl() })
        else:
            proxy_handler = urllib2.ProxyHandler({})
            opener = urllib2.build_opener(proxy_handler)
            urllib2.install_opener(opener)
            #os.environ['http_proxy']=''
            #self.proxy = urllib2.ProxyHandler()
        #self.opener = urllib2.urlopen #build_opener(self.proxy)
        #urllib2.install_opener(self.opener)

    def test(self):
        '''Test downloading the pUC19 plasmid sequence from genbank
        returns True if successful. Can be used to test proxy
        and other network settings.
        '''

        try:
            result = self.nucleotide("L09137") # pUC19
        except urllib2.URLError:
            return False
        return seguid(result.seq) == "71B4PwSgBZ3htFjJXwHPxtUIPYE"

    def nucleotide(self, item, start=None, stop=None, strand="watson" ):
        '''Download a genbank nuclotide record. 

       Item is a string containing one genbank acession number [#]_
       for a nucleotide file. Start and stop are intervals to be
       downloaded. This is useful as some genbank records are large.
       If strand is "c", "C", "crick", "Crick", "antisense","Antisense", 
       "2" or 2, the watson(antisense) strand is returned, otherwise
       the sense strand is returned.
       
       Alternatively, item can be a string containing an url that returns a 
       sequence in genbank or FASTA format.
       
       The genbank E-utilities can provide such urls [#]_.
       
       Result is returned as a Dseqrecord object.
       
       Genbank nucleotide accession numbers have this format:

       | A12345   = 1 letter  + 5 numerals
       | AB123456 = 2 letters + 6 numerals
       

       References
       ----------

       .. [#]   http://www.dsimb.inserm.fr/~fuchs/M2BI/AnalSeq/Annexes/Sequences/Accession_Numbers.htm
       .. [#]   http://www.ncbi.nlm.nih.gov/books/NBK25499/#chapter4.EFetch
       '''
       
        if item.lower().startswith(("ftp","http")):
            try:
                url = urlparse(item)
            except (AttributeError, TypeError):
                url = urlparse("")

            if url.scheme:
                return read( urllib2.urlopen(item).read() )
        
        matches =((1, re.search("(REGION:\s(?P<start>\d+)\.\.(?P<stop>\d+))", item)),
                  (2, re.search("(REGION: complement\((?P<start>\d+)\.\.(?P<stop>\d+)\))",item)),
                  (1, re.search(":(?P<start>\d+)-(?P<stop>\d+)",item)),
                  (2, re.search(":c(?P<start>\d+)-(?P<stop>\d+)",item)),
                  (0, None))

        # BK006936.2 REGION: complement(613900..615202)
        # NM_005546 REGION: 1..100
        # NM_005546 REGION: complement(1..100)
        # 21614549:1-100
        # 21614549:c100-1
        
        # http://eutils.ncbi.nlm.nih.gov/entrez/eutils/efetch.fcgi?db=nuccore&id=BK006936.2&strand=2&seq_start=613900&seq_stop=615202&rettype=fasta&retmode=text
        # http://eutils.ncbi.nlm.nih.gov/entrez/eutils/efetch.fcgi?db=nuccore&id=21614549&strand=1&seq_start=1&seq_stop=100&rettype=fasta&retmode=text
        # http://eutils.ncbi.nlm.nih.gov/entrez/eutils/efetch.fcgi?db=nuccore&id=21614549&strand=2&seq_start=1&seq_stop=100&rettype=fasta&retmode=text
        # http://eutils.ncbi.nlm.nih.gov/entrez/eutils/efetch.fcgi?db=nuccore&id=21614549&strand=1&seq_start=1&seq_stop=100&rettype=gb&retmode=text
        # http://eutils.ncbi.nlm.nih.gov/entrez/eutils/efetch.fcgi?db=nuccore&id=21614549&strand=2&seq_start=1&seq_stop=100&rettype=gb&retmode=text
        

        for strand, match in matches:
            if match:
                start = match.group("start")
                stop  = match.group("stop")
                item = item[:match.start()]
                break
            
    #        if not strand:
    #            raise Exception("\nACESSION string is malformed!\n"
    #                              "NM_005546 REGION: 1..100\n"
    #                              "NM_005546 REGION: complement(1..100)\n"
    #                              "21614549:1-100\n"
    #                              "21614549:c100-1\n")

        if str(strand).lower() in ("c","crick", "antisense", "2"):
            strand = 2
        else:
            strand = 1
            
        Entrez.email = self.email
        
#        print item
#        print start
#        print stop
#        print strand

        return read(Entrez.efetch(db        ="nucleotide",
                                  id        = item,
                                  rettype   = "gb",
                                  seq_start = start,
                                  seq_stop  = stop,
                                  strand    = strand,
                                  retmode   = "text").read())
        


if __name__=="__main__":
    import doctest
    doctest.testmod()
    gb = Genbank("bjornjobbb@gmail.com")
    #print gb.email
    

    gene = gb.nucleotide("BK006936.2 REGION: complement(613900..615202)")
    



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# doctest: +NORMALIZE_WHITESPACE
# doctest: +SKIP
'''This module provides functions for PCR. Primers with 5' tails as well as inverse PCR on
circular templates are handled correctly.

'''

import itertools
import math
import re
import textwrap
import copy
import operator

from math                           import log10
from Bio.Seq                        import Seq
from Bio.Alphabet.IUPAC             import ambiguous_dna
from Bio.SeqRecord                  import SeqRecord
from Bio.SeqUtils                   import GC
from Bio.SeqUtils.MeltingTemp       import Tm_staluc
from Bio.SeqFeature                 import SeqFeature
from Bio.SeqFeature                 import CompoundLocation
from Bio.SeqFeature                 import FeatureLocation
from pydna.dsdna                    import Dseq
from pydna.dsdna                    import rc
from pydna.dsdna                    import parse
from pydna.dsdna                    import Dseqrecord

def _annealing_positions(primer, template, limit=15):
    '''Finds the annealing position(s) for a primer on a template where the
    primer anneals perfectly with at least limit nucleotides in the 3' part.

    start is a position (integer)
    footprint1 and tail1 are strings.

    ::

        <- - - - - - - - - - template - - - - - - - - - - - - - >

        <------------- start ---->
     5'-...gctactacacacgtactgactgcctccaagatagagtcagtaaccacactcgat...3'
           ||||||||||||||||||||||||||||||||||||||||||||||||
                                  3'-gttctatctcagtcattggtgtATAGTG-5'

                                                        <tail>
                                     <---footprint----->
                                     <--------- primer ------>

    Parameters
    ----------
    primer : string
        The primer sequence 5'-3'

    template : string
        The template sequence 5'-3'

    limit : int = 15, optional
        footprint needs to be at least of length limit.

    Returns
    -------
    describe : list of tuples (int, string, string)
        [ (start1, footprint1, tail1), (start2, footprint2, tail2),..., ]
    '''

    if len(primer)<limit:
        return []
    prc = rc(primer)
    head = prc[:limit].upper()

    table = {"R":"(A|G)",
             "Y":"(C|T)",
             "S":"(G|C)",
             "W":"(A|T)",
             "K":"(G|T)",
             "M":"(A|C)",
             "B":"(C|G|T)",
             "D":"(A|G|T)",
             "H":"(A|C|T)",
             "V":"(A|C|G)",
             "N":"(A|G|C|T)"}

    for key in table:
        head=head.replace(key, table[key])

    positions = [m.start() for m in re.finditer('(?={})'.format(head), template, re.I)]
    if positions:
        tail = prc[limit:]
        length = len(tail)
        #revtail = tail[::-1]
        results = []
        for match_start in positions:
            tm = template[match_start+limit:match_start+limit+length]
            footprint = rc(template[match_start:match_start+limit]+"".join([b for a,b in itertools.takewhile(lambda x: x[0].lower()==x[1].lower(),zip(tail, tm))]))
            results.append((match_start, footprint, primer[: len(primer) - len(footprint) ]))
        return results
    return []

class Primer(SeqRecord):
    '''This class holds information about a primer on a template '''

    def __init__(self,
                 seq_obj,
                 position,
                 footprint,
                 tail):

        self.position  = position
        self.footprint = footprint
        self.tail      = tail

        seq_obj.seq.alphabet = ambiguous_dna

        SeqRecord.__init__(self,
                           seq=seq_obj.seq,
                           id=seq_obj.id,
                           name=seq_obj.name,
                           description=seq_obj.description,
                           dbxrefs=seq_obj.dbxrefs,
                           features=seq_obj.features,
                           annotations=seq_obj.annotations,
                           letter_annotations=seq_obj.letter_annotations)


class Amplicon(Dseqrecord):
    '''The Amplicon class holds information about a PCR reaction involving two
    primers and one template. This class is used by the Anneal class and is not
    meant to be instantiated directly.

    Parameters
    ----------
    forward_primer : SeqRecord(Biopython)
        SeqRecord object holding the forward (sense) primer

    reverse_primer : SeqRecord(Biopython)
        SeqRecord object holding the reverse (antisense) primer

    template : Dseqrecord
        Dseqrecord object holding the template (circular or linear)

    saltc : float, optional
        saltc = monovalent cations (mM) (Na,K..)
        default value is 50mM
        This is used for Tm calculations.

    forward_primer_concentration : float, optional
        primer concentration (nM)
        default set to 1000nM = 1µM
        This is used for Tm calculations.

    rc : float, optional
        primer concentration (nM)
        default set to 1000nM = 1µM
        This is used for Tm calculations.
    '''

    def __init__(    self,
                     record,
                     template=None,
                     forward_primer=None,
                     reverse_primer=None,
                     saltc=None,
                     forward_primer_concentration=None,
                     reverse_primer_concentration=None,
                     *args,
                     **kwargs):

        #Dseqrecord.__init__(self,record,*args,**kwargs)
        super(Amplicon, self).__init__(record, *args, **kwargs)
        self.template = template
        self.forward_primer = forward_primer
        self.reverse_primer = reverse_primer
        self.forward_primer_concentration = forward_primer_concentration
        self.reverse_primer_concentration = reverse_primer_concentration
        self.saltc = saltc

        self.tmf = Tm_staluc(str(self.forward_primer.footprint),dnac=50, saltc=self.saltc)
        self.tmr = Tm_staluc(str(self.reverse_primer.footprint),dnac=50, saltc=self.saltc)
        self.tmf_dbd = tmbresluc(str(self.forward_primer.footprint),primerc=self.forward_primer_concentration)
        self.tmr_dbd = tmbresluc(str(self.reverse_primer.footprint),primerc=self.reverse_primer_concentration)


    def __getitem__(self, sl):
        answer = copy.copy(self)
        answer.seq = answer.seq.__getitem__(sl)
        answer.seq.alphabet = self.seq.alphabet
        sr = SeqRecord("n"*len(self))
        sr.features = self.features
        answer.features = SeqRecord.__getitem__(sr, sl).features
        return answer

    def __repr__(self):
        '''returns a short string representation of the object'''
        return "Amplicon({})".format(self.__len__())

    def flankup(self, flankuplength=50):
        '''Returns a Dseqrecord object containing flankuplength bases upstream of the forward primer footprint,
       Truncated if the template is not long enough.

       ::

        <--- flankup --->

                  5TAATAAactactgactatct3
                         ||||||||||||||
        acgcattcagctactgtactactgactatctatcg

       '''
        return self.template.seq[self.forward_primer.position-flankuplength-len(self.forward_primer.footprint):self.forward_primer.position-len(self.forward_primer.footprint)]

    def flankdn(self, flankdnlength=50):
        '''Returns a Dseqrecord object containing flankdnlength bases downstream of the reverse primer footprint.
       Truncated if the template is not long enough.

       ::

                                       <---- flankdn ------>

                        3actactgactatctTAATAA5
                         ||||||||||||||
        acgcattcagctactgtactactgactatctatcgtacatgtactatcgtat


       '''
        return self.template.seq[self.reverse_primer.position+len(self.reverse_primer.footprint):self.reverse_primer.position+flankdnlength+len(self.reverse_primer.footprint)]



    def figure(self):
        '''
        This method returns a simple figure of the two primers binding to a part
        of the template.

        ::

         5gctactacacacgtactgactg3
          |||||||||||||||||||||| tm 52.6 (dbd) 58.3
         5gctactacacacgtactgactg...caagatagagtcagtaaccaca3
         3cgatgatgtgtgcatgactgac...gttctatctcagtcattggtgt5
                                   |||||||||||||||||||||| tm 49.1 (dbd) 57.7
                                  3gttctatctcagtcattggtgt5



        Returns
        -------
        figure:string
             A string containing a text representation of the primers
             annealing on the template (see example above).


        Notes
        -----
        tm in the figure above is the melting temperature (tm) for each primer calculated according to
        SantaLucia 1998 [#]_.

        dbd is the tm calculation for enzymes with dsDNA binding domains like Pfu-Sso7d [#]_. See [#]_
        for more information.

        References
        ----------

        .. [#] J. SantaLucia, “A Unified View of Polymer, Dumbbell, and Oligonucleotide DNA Nearest-neighbor Thermodynamics,” Proceedings of the National Academy of Sciences 95, no. 4 (1998): 1460.
        .. [#] M. Nørholm, “A Mutant Pfu DNA Polymerprimerase Designed for Advanced Uracil-excision DNA Engineering,” BMC Biotechnology 10, no. 1 (2010): 21, doi:10.1186/1472-6750-10-21.
        .. [#] http://www.thermoscientificbio.com/webtools/tmc/

        '''

        f =   '''
            {sp1}5{faz}...{raz}3
             {sp3}{rap} tm {tmr} (dbd) {tmr_dbd}
            {sp3}3{rp}5
            5{fp}3
             {fap:>{fplength}} tm {tmf} (dbd) {tmf_dbd}
            {sp2}3{fzc}...{rzc}5
            '''.format( fp       = self.forward_primer.seq,
                        fap      = "|"*len(self.forward_primer.footprint),
                        fplength = len(self.forward_primer.seq),
                        tmf      = round(self.tmf,1),
                        tmr      = round(self.tmr,1),
                        tmf_dbd  = round(self.tmf_dbd,1),
                        tmr_dbd  = round(self.tmr_dbd,1),
                        rp       = self.reverse_primer.seq[::-1],
                        rap      = "|"*len(self.reverse_primer.footprint),
                        rplength = len(self.reverse_primer.seq),
                        faz      = self.forward_primer.footprint,
                        raz      = self.reverse_primer.footprint.reverse_complement(),
                        fzc      = self.forward_primer.footprint.complement(),
                        rzc      = self.reverse_primer.footprint[::-1],
                        sp1      = " "*(len(self.forward_primer.seq)-len(self.forward_primer.footprint)),
                        sp2      = " "*(len(self.forward_primer.seq)-len(self.forward_primer.footprint)),
                        sp3      = " "*(3+len(self.forward_primer.seq))
                       )
        return textwrap.dedent(f).strip("\n")

    def program(self):
        '''Returns a string containing a text representation of two proposed
       PCR programs. The first program is adapted for Taq poymerase, while
       the second is adapted for Pfu-Sso7d or other polymerases with DNA binding domain.

       ::

        Taq (rate 30 nt/s)
        Three-step|         30 cycles     |      |SantaLucia 1998
        94.0°C    |94.0°C                 |      |SaltC 50mM
        __________|_____          72.0°C  |72.0°C|
        04min00s  |30s  \         ________|______|
                  |      \ 46.0°C/ 0min 1s|10min |
                  |       \_____/         |      |
                  |         30s           |      |4-8°C

        Pfu-Sso7d (rate 15s/kb)
        Three-step|          30 cycles   |      |Breslauer1986,SantaLucia1998
        98.0°C    |98.0°C                |      |SaltC 50mM
        __________|_____          72.0°C |72.0°C|Primer1C   1µM
        00min30s  |10s  \ 61.0°C ________|______|Primer2C   1µM
                  |      \______/ 0min 0s|10min |
                  |        10s           |      |4-8°C

       '''

        # Ta calculation according to
        # Rychlik, Spencer, and Rhoads, 1990, Optimization of the anneal
        # ing temperature for DNA amplification in vitro
        # http://www.ncbi.nlm.nih.gov/pubmed/2003928
        GC_prod=GC(str(self.seq))
        tml = min(self.tmf,self.tmr)
        #print GC(str(self.product.seq)), self.saltc/1000.0, len(self.product)
        tmp = 81.5 + 0.41*GC(str(self.seq)) + 16.6*log10(self.saltc/1000.0) - 675/len(self)
        ta = 0.3*tml+0.7*tmp-14.9
        # Fermentas recombinant taq
        taq_extension_rate = 30  # seconds/kB PCR product length
        extension_time_taq = taq_extension_rate * len(self) / 1000 # seconds
        f  = textwrap.dedent( '''
                                Taq (rate {rate} nt/s)
                                Three-step|         30 cycles     |      |SantaLucia 1998
                                94.0°C    |94.0°C                 |      |SaltC {saltc:2}mM
                                __________|_____          72.0°C  |72.0°C|
                                04min00s  |30s  \         ________|______|
                                          |      \ {ta}°C/{0:2}min{1:2}s|10min |
                                          |       \_____/         |      |
                                          |         30s           |      |4-8°C
                             '''.format(rate    = taq_extension_rate,
                                        ta      = math.ceil(ta),
                                        saltc   = self.saltc,
                                        *divmod(extension_time_taq,60)))

        PfuSso7d_extension_rate = 15 #seconds/kB PCR product
        extension_time_PfuSso7d = PfuSso7d_extension_rate * len(self) / 1000  # seconds

        # Ta calculation for enzymes with dsDNA binding domains like Pfu-Sso7d
        # https://www.finnzymes.fi/tm_determination.html

        length_of_f = len(self.forward_primer.footprint)
        length_of_r = len(self.reverse_primer.footprint)

        if (length_of_f>20 and length_of_r>20 and self.tmf_dbd>=69.0 and self.tmr_dbd>=69.0) or (self.tmf_dbd>=72.0 and self.tmr_dbd>=72.0):
            f+=textwrap.dedent(  '''
                                    Pfu-Sso7d (rate {rate}s/kb)
                                    Two-step|    30 cycles |      |Breslauer1986,SantaLucia1998
                                    98.0°C  |98.0C         |      |SaltC {saltc:2}mM
                                    _____ __|_____         |      |Primer1C {forward_primer_concentration:3}µM
                                    00min30s|10s  \  72.0°C|72.0°C|Primer2C {reverse_primer_concentration:3}µM
                                            |      \_______|______|
                                            |      {0:2}min{1:2}s|10min |4-8°C
                                 '''.format(rate = PfuSso7d_extension_rate,
                                            forward_primer_concentration = self.forward_primer_concentration,
                                            reverse_primer_concentration = self.rc,
                                            saltc = self.saltc,
                                            *divmod(extension_time_PfuSso7d,60)))
        else:

            if (length_of_f>20 and length_of_r>20):
                ta = min(self.tmf_dbd,self.tmr_dbd)+3
            else:
                ta = min(self.tmf_dbd,self.tmr_dbd)


            f+=textwrap.dedent(  '''
                                    Pfu-Sso7d (rate {rate}s/kb)
                                    Three-step|          30 cycles   |      |Breslauer1986,SantaLucia1998
                                    98.0°C    |98.0°C                |      |SaltC {saltc:2}mM
                                    __________|_____          72.0°C |72.0°C|Primer1C {forward_primer_concentration:3}µM
                                    00min30s  |10s  \ {ta}°C ________|______|Primer2C {reverse_primer_concentration:3}µM
                                              |      \______/{0:2}min{1:2}s|10min |
                                              |        10s           |      |4-8°C
                                 '''.format(rate = PfuSso7d_extension_rate,
                                            ta   = math.ceil(ta),
                                            forward_primer_concentration   = self.forward_primer_concentration/1000,
                                            reverse_primer_concentration   = self.reverse_primer_concentration/1000,
                                            saltc= self.saltc,
                                            *divmod(extension_time_PfuSso7d,60)))
        return f



class Anneal(object):
    '''

    Parameters
    ----------
    primers : iterable containing SeqRecord objects
        Primer sequences 5'-3'.

    template : Dseqrecord object
        The template sequence 5'-3'.

    limit : int, optional
        limit length of the annealing part of the primers.

    Attributes
    ----------
    products: list
        A list of Amplicon objects, one for each primer pair that may form a PCR product.


    Examples
    --------
    >>> import pydna
    >>> template = pydna.Dseqrecord("tacactcaccgtctatcattatctactatcgactgtatcatctgatagcac")
    >>> from Bio.SeqRecord import SeqRecord
    >>> p1 = pydna.read(">p1\\ntacactcaccgtctatcattatc", ds = False)
    >>> p2 = pydna.read(">p2\\ngtgctatcagatgatacagtcg", ds = False)
    >>> ann = pydna.Anneal((p1,p2), template)
    >>> print ann.report()
    Template <unknown name> 51 nt linear:
    Primer p1 anneals forward at position 23
    Primer p2 anneals reverse at position 29
    >>> ann.products
    [Amplicon(51)]
    >>> amplicon_list = ann.products
    >>> amplicon = amplicon_list.pop()
    >>> amplicon
    Amplicon(51)
    >>> print amplicon.figure()
    5tacactcaccgtctatcattatc...cgactgtatcatctgatagcac3
                               |||||||||||||||||||||| tm 50.6 (dbd) 60.5
                              3gctgacatagtagactatcgtg5
    5tacactcaccgtctatcattatc3
     ||||||||||||||||||||||| tm 49.4 (dbd) 58.8
    3atgtgagtggcagatagtaatag...gctgacatagtagactatcgtg5
    >>> amplicon.annotations['date'] = '02-FEB-2013'   # Set the date for this example to pass the doctest
    >>> print amplicon
    Dseqrecord
    circular: False
    size: 51
    ID: 51bp U96+TO06Y6pFs74SQx8M1IVTBiY
    Name: 51bp_PCR_prod
    Description: Product_p1_p2
    Number of features: 2
    /date=02-FEB-2013
    Dseq(-51)
    taca..gcac
    atgt..cgtg
    >>> print amplicon.program()
    <BLANKLINE>
    Taq (rate 30 nt/s)
    Three-step|         30 cycles     |      |SantaLucia 1998
    94.0°C    |94.0°C                 |      |SaltC 50mM
    __________|_____          72.0°C  |72.0°C|
    04min00s  |30s  \         ________|______|
              |      \ 44.0°C/ 0min 1s|10min |
              |       \_____/         |      |
              |         30s           |      |4-8°C
    <BLANKLINE>
    Pfu-Sso7d (rate 15s/kb)
    Three-step|          30 cycles   |      |Breslauer1986,SantaLucia1998
    98.0°C    |98.0°C                |      |SaltC 50mM
    __________|_____          72.0°C |72.0°C|Primer1C   1µM
    00min30s  |10s  \ 62.0°C ________|______|Primer2C   1µM
              |      \______/ 0min 0s|10min |
              |        10s           |      |4-8°C
    <BLANKLINE>
    >>>

    '''
    def __init__( self,
                  primers,
                  template,
                  limit=13):

        self.template = copy.deepcopy(template)
        #self.template.features = copy.deepcopy(template.features)

        self.limit = limit
        self._products = None

        primers=[p for p in primers if p.seq]

        self.fwd_primers = []
        self.rev_primers = []

        twl = len(self.template.seq.watson)
        tcl = len(self.template.seq.crick)

        if self.template.linear:
            tw = self.template.seq.watson
            tc = self.template.seq.crick
        else:
            tw = self.template.seq.watson+self.template.seq.watson
            tc = self.template.seq.crick +self.template.seq.crick

        for p in primers:
            self.fwd_primers.extend((Primer(p,
                                            tcl-pos - min(self.template.seq.ovhg, 0),
                                            Seq(fp), Seq(tl))
                                    for pos, fp, tl in _annealing_positions(
                                                        str(p.seq),
                                                        tc,
                                                        limit) if pos<tcl))
            self.rev_primers.extend((Primer(p,
                                            pos + max(0, self.template.seq.ovhg),
                                            Seq(fp), Seq(tl))
                                     for pos, fp, tl in _annealing_positions(
                                                                     str(p.seq),
                                                                     tw,
                                                                     limit) if pos<twl))
        self.fwd_primers.sort(key = operator.attrgetter('position'))
        self.rev_primers.sort(key = operator.attrgetter('position'), reverse=True)


        for fp in self.fwd_primers:
            if fp.position-len(fp.footprint)>=0:
                start = fp.position-len(fp.footprint)
                end   = fp.position
                self.template.features.append(SeqFeature(FeatureLocation(start, end),
                                                    type ="primer_bind",
                                                    strand = 1,
                                                    qualifiers = {"note":[fp.name],
                                                                  "ApEinfo_fwdcolor":["green"],
                                                                  "ApEinfo_revcolor":["red"]}))
            else:
                start = len(self.template)-len(fp.footprint)+fp.position
                end = start+len(fp.footprint)-len(self.template)
                sf=SeqFeature(CompoundLocation([FeatureLocation(start,len(self.template)),
                                                FeatureLocation(0, end)]),
                                                type="primer_bind",
                                                location_operator="join",
                                                qualifiers = {"note":[fp.name]})

                self.template.features.append(sf)

        for rp in self.rev_primers:
            if rp.position+len(rp.footprint)<=len(self.template):
                start = rp.position
                end   = rp.position + len(rp.footprint)
                self.template.features.append(SeqFeature(FeatureLocation(start,end),
                                                    type ="primer_bind",
                                                    strand = -1,
                                                    qualifiers = {"note":[rp.name],
                                                                  "ApEinfo_fwdcolor":["green"],
                                                                  "ApEinfo_revcolor":["red"]}))
            else:
                start = rp.position
                end = rp.position+len(rp.footprint)-len(self.template)
                self.template.features.append(SeqFeature(CompoundLocation([FeatureLocation(start,len(self.template)),
                                                                           FeatureLocation(0,end)]),
                                                    type ="primer_bind",
                                                    location_operator= "join",
                                                    strand = -1,
                                                    qualifiers = {"note":[rp.name]}))


    @property
    def products(self):

        if self._products:
            return self._products

        self._products = []

        for fp in self.fwd_primers:
            for rp in self.rev_primers:
                if self.template.circular:
                    tmpl=self.template._multiply_circular(2)
                else:
                    tmpl=self.template

                prd = ( Dseqrecord(fp.tail) + Dseqrecord(fp.footprint) +
                        tmpl[fp.position : rp.position+len(self.template) if fp.position>rp.position else rp.position] +
                        Dseqrecord(rp.footprint.reverse_complement()) + Dseqrecord(rp.tail.reverse_complement())
                        )

                features = tmpl[fp.position-len(fp.footprint):rp.position+len(rp.footprint)].features

                prd.features = [f._shift(len(fp.tail)) for f in features]

                # description = Genbank LOCUS max 16 chars

                prd.name = "{0}bp_PCR_prod".format(len(prd))[:16]
                prd.id = "{0}bp {1}".format( str(len(prd))[:14], prd.seguid() )
                prd.description="Product_{0}_{1}".format( fp.description,
                                                          rp.description)

                self._products.append( Amplicon(prd,
                                                template=tmpl,
                                                forward_primer=fp,
                                                reverse_primer=rp,
                                                saltc=50,
                                                forward_primer_concentration=1000,
                                                reverse_primer_concentration=1000))
        return self._products



    def report(self):
        '''This method is an alias of str(Annealobj).
        Returns a short string representation.
        '''
        return self.__str__()

    def __repr__(self):
        ''' returns a short string represenation '''
        return "Reaction(products = {})".format(len(self.fwd_primers*len(self.rev_primers)))

    def __str__(self):
        '''returns a short report describing if or where primer
       anneal on the template.
       '''
        mystring = "Template {name} {size} nt {top}:\n".format(name=self.template.name,
                                                               size=len(self.template),
                                                               top={True:"circular",
                                                                    False:"linear"}[self.template.circular]
                                                                    )
        if self.fwd_primers:
            for p in self.fwd_primers:
                mystring += "Primer {name} anneals forward at position {pos}\n".format(name=p.name, pos=p.position)
        else:
            mystring += "No forward primers anneal...\n"
        if self.rev_primers:
            for p in self.rev_primers:
                mystring += "Primer {name} anneals reverse at position {pos}\n".format(name=p.name, pos=p.position)
        else:
             mystring += "No reverse primers anneal...\n"
        return mystring.strip()






def pcr(*args,  **kwargs):
    '''pcr is a convenience function for Anneal to simplify its usage,
    especially from the command line. If more than one PCR product is
    formed, an exception is raised.

    args is any iterable of sequences or an iterable of iterables of sequences.
    args will be greedily flattened.

    Parameters
    ----------

    args : iterable containing sequence objects
        Several arguments are also accepted.

    limit : int = 13, optional
        limit length of the annealing part of the primers.

    Notes
    -----

    sequences in args could be of type:

    * string
    * Seq
    * SeqRecord
    * Dseqrecord

    The last sequence will be interpreted as the template
    all preceeding sequences as primers.

    This is a powerful function, use with care!

    Returns
    -------

    product : Dseqrecord
        a Dseqrecord object representing the PCR product.
        The direction of the PCR product will be the same as
        for the template sequence.

    Examples
    --------

    >>> import pydna
    >>> template = pydna.Dseqrecord("tacactcaccgtctatcattatctactatcgactgtatcatctgatagcac")
    >>> from Bio.SeqRecord import SeqRecord
    >>> p1 = pydna.read(">p1\\ntacactcaccgtctatcattatc", ds = False)
    >>> p2 = pydna.read(">p2\\ncgactgtatcatctgatagcac", ds = False).reverse_complement()
    >>> pydna.pcr(p1, p2, template)
    Amplicon(51)
    >>> pydna.pcr([p1, p2], template)
    Amplicon(51)
    >>> pydna.pcr((p1,p2,), template)
    Amplicon(51)
    >>>

    '''

    from Bio.SeqRecord import SeqRecord
    # flatten args
    output = []
    stack = []
    stack.extend(reversed(args))
    while stack:
        top = stack.pop()
        if hasattr(top, "__iter__") and not isinstance(top, SeqRecord):
            stack.extend(reversed(top))
        else:
            output.append(top)
    new=[]

    for s in output:
        if isinstance(s, Seq):
            s = SeqRecord(s)
        elif isinstance(s, SeqRecord):
            pass
        elif hasattr(s, "watson"):
            s=s.watson
        elif isinstance(s, basestring):
            s = SeqRecord(Seq(s))
        else:
            raise TypeError("the record property needs to be a string, a Seq object or a SeqRecord object")
        new.append(s)

    anneal_primers = Anneal(  new[:-1],
                              new[-1],
                              **kwargs)

    if anneal_primers:
        if len(anneal_primers.products) == 1:
            return anneal_primers.products.pop()
        elif len(anneal_primers.products) == 0:
            raise Exception("No PCR products! {}".format(anneal_primers.report()))
        else:
            raise Exception("PCR not specific! {}".format(anneal_primers.report()))
    else:
        raise Exception(anneal_primers.report())
    return


def basictm(primer, *args, **kwargs):
    '''Returns the melting temperature (Tm) of the primer using
    the basic formula.

    | Tm = (wA+xT)*2 + (yG+zC)*4 assumed 50mM monovalent cations
    |
    | w = number of A in primer
    | x = number of T in primer
    | y = number of G in primer
    | z = number of C in primer

    Parameters
    ----------
    primer : string
        Primer sequence 5'-3'

    Returns
    -------
    tm : int

    Examples
    --------
    >>> from pydna.amplify import basictm
    >>> basictm("ggatcc")
    20
    >>>

    '''
    primer = str(primer).lower()
    return (primer.count("a") + primer.count("t"))*2 + (primer.count("g") + primer.count("c"))*4

# http://www.promega.com/techserv/tools/biomath/calc11.htm#melt_results

def tmstaluc98(primer, dnac=50, saltc=50, **kwargs):
    '''Returns the melting temperature (Tm) of the primer using
    the nearest neighbour algorithm. Formula and thermodynamic data
    is taken from SantaLucia 1998 [1]_. This implementation gives the same
    answer as the one provided by Biopython (See Examples).

    Thermodynamic data used:

    =====  ====  ====
    pair   dH    dS
    =====  ====  ====
    AA/TT  7.9   22.2
    AT/TA  7.2   20.4
    TA/AT  7.2   21.3
    CA/GT  8.5   22.7
    GT/CA  8.4   22.4
    CT/GA  7.8   21.0
    GA/CT  8.2   22.2
    CG/GC  10.6  27.2
    GC/CG  9.8   24.4
    GG/CC  8.0   19.9
    =====  ====  ====

    Parameters
    ----------
    primer : string
        Primer sequence 5'-3' in UPPERCASE

    Returns
    -------
    tm : float
        tm of the primer


    Examples
    --------

    >>> from pydna.amplify import tmstaluc98
    >>> from Bio.SeqUtils.MeltingTemp import Tm_staluc
    >>> tmstaluc98("ACGTCATCGACACTATCATCGAC")
    54.55597724052518
    >>> Tm_staluc("ACGTCATCGACACTATCATCGAC")
    54.555977240525124
    >>>



    '''

    nntermsl={  "AA": (7.9  , 22.2),
                "TT": (7.9  , 22.2),
                "AT": (7.2  , 20.4),
                "TA": (7.2  , 21.3),
                "CA": (8.5  , 22.7),
                "TG": (8.5  , 22.7),
                "GT": (8.4  , 22.4),
                "AC": (8.4  , 22.4),
                "CT": (7.8  , 21.0),
                "AG": (7.8  , 21.0),
                "GA": (8.2  , 22.2),
                "TC": (8.2  , 22.2),
                "CG": (10.6 , 27.2),
                "GC": (9.8  , 24.4),
                "GG": (8    , 19.9),
                "CC": (8    , 19.9),
                "A" : (0    , 0   ),
                "C" : (0    , 0   ),
                "G" : (0    , 0   ),
                "T" : (0    , 0   )  }

    helixinit = {   "G": (-0.1 ,2.8),
                    "C": (-0.1 ,2.8),
                    "A": (-2.3, -4.1),
                    "T": (-2.3, -4.1) }

    dH, dS = helixinit[primer[0]]
    H ,  S = helixinit[primer[-1]]
    dH = dH+H
    dS = dS+S

    for p in range(len(primer)):
        dn = primer[p:p+2]
        H,S = nntermsl[dn]
        dH+=H
        dS+=S
    R = 1.987 # universal gas constant in Cal/degrees C*Mol
    k = (dnac/4.0)*1e-9
    dS = dS-0.368*(len(primer)-1)*math.log(float(saltc)/1e3)
    tm = ((1000* (-dH))/(-dS+(R * (math.log(k)))))-273.15
    return tm

def tmbreslauer86(primer, dnac=500.0, saltc=50,thermodynamics=False):
    '''Returns the melting temperature (Tm) of the primer using
    the nearest neighbour algorithm. Formula and thermodynamic data
    is taken from Breslauer 1986. These data are no longer widely used.


    Breslauer 1986, table 2 [#]_

    =====  ===== ====   ===
    pair   dH    dS     dG
    =====  ===== ====   ===
    AA/TT  9.1   24.0   1.9
    AT/TA  8.6   23.9   1.5
    TA/AT  6.0   16.9   0.9
    CA/GT  5.8   12.9   1.9
    GT/CA  6.5   17.3   1.3
    CT/GA  7.8   20.8   1.6
    GA/CT  5.6   13.5   1.6
    CG/GC  11.9  27.8   3.6
    GC/CG  11.1  26.7   3.1
    GG/CC  11.0  26.6   3.1
    =====  ===== ====   ===

    Parameters
    ----------
    primer : string
        Primer sequence 5'-3' in UPPERCASE

    Returns
    -------
    tm : float


    References
    ----------
    .. [#] K.J. Breslauer et al., “Predicting DNA Duplex Stability from the Base Sequence,” Proceedings of the National Academy of Sciences 83, no. 11 (1986): 3746.


    Examples
    ---------

    >>> from pydna.amplify import tmbreslauer86
    >>> tmbreslauer86("ACGTCATCGACACTATCATCGAC")
    64.28863985851899


    '''

    nntermbr={  "AA": (9.1   ,24.0   ,1.9),
                "TT": (9.1   ,24.0   ,1.9),
                "AT": (8.6   ,23.9   ,1.5),
                "TA": (6.0   ,16.9   ,0.9),
                "CA": (5.8   ,12.9   ,1.9),
                "TG": (5.8   ,12.9   ,1.9),
                "GT": (6.5   ,17.3   ,1.3),
                "AC": (6.5   ,17.3   ,1.3),
                "CT": (7.8   ,20.8   ,1.6),
                "AG": (7.8   ,20.8   ,1.6),
                "GA": (5.6   ,13.5   ,1.6),
                "TC": (5.6   ,13.5   ,1.6),
                "CG": (11.9  ,27.8   ,3.6),
                "GC": (11.1  ,26.7   ,3.1),
                "GG": (11.0  ,26.6   ,3.1),
                "CC": (11.0  ,26.6   ,3.1),
                "A" : (0     , 0     ,0),
                "C" : (0     , 0     ,0),
                "G" : (0     , 0     ,0),
                "T" : (0     , 0     ,0),     }
    dH=3.4
    dS=12.4
    dG=0
    for p in range(len(primer)):
        dn = primer[p:p+2]
        H,S,G = nntermbr[dn]
        dG+=G
        dH+=H
        dS+=S

    R = 1.9872          # universal gas constant in Cal/degrees C*Mol
    k = dnac*1E-9/2.0
    dH = dH - 5
    dS = dS-0.368*(len(primer)-1)*math.log(float(saltc)/1E3) # SantaLucia salt correction formula
    tm = 1000 * -dH /(-dS + R * math.log(k) )  - 273.15 # degrees Celsius

    if thermodynamics:
        return tm,dH,dS
    else:
        return tm


def tmbresluc(primer, primerc=500.0, saltc=50, thermodynamics=False):
    '''Returns the tm for a primer using a formula adapted to polymerases
    with a DNA binding domain.

    Parameters
    ----------

    primer : string
        primer sequence 5'-3'

    primerc : float
       concentration (nM)

    saltc : float, optional
       Monovalent cation concentration (mM)

    thermodynamics : bool, optional
        prints details of the thermodynamic data to stdout. For
        debugging only.

    Returns
    -------
    tm : float
        the tm of the primer

    tm,dH,dS : tuple
        tm and dH and dS used for the calculation

    '''

    import thermodynamic_data

    saltc = float(saltc)/1000
    pri  = primerc/10E7
    dS = -12.4
    dH = -3400

    STR = primer.lower();

    for i in range(len(STR)-1):
        n1=ord(STR[i])
        n2=ord(STR[i+1])
        dH += thermodynamic_data.dHBr[n1 - 97][n2 - 97]
        dS += thermodynamic_data.dSBr[n1 - 97][n2 - 97]

    tm = (dH / (1.9872 * math.log(pri / 1600) + dS) + (16.6 * math.log(saltc)) / math.log(10)) - 273.15

    if thermodynamics:
        return tm,dH,dS
    else:
        return tm



if __name__=="__main__":
    import doctest
    doctest.testmod()

    import pydna
    f, r = pydna.parse('''
    >f
    ARATGAGTCTTCTRACCGAGGTCG
    >r
    TGAAAAGACATCYTCAAGYYTCTG''', ds=False)

    t= pydna.read('''>templ
    AGCAAAAGCAGGTAGATATTGAAAAATGAGTCTTCTAACCGAGGTCGAAACGTACGTTCTCTCTATCGTCCCGTCAGGCCCCCTCAAAGCCGAGATCGCGCAGAGACTTGAAGATGTCTCTGCAGGGAAGAACACTGATCTCGAGGCTCTCATGGAATGGCTAAAGACAAGACCAATCCTGTCACCTCTGACTAAGGGGATTTTAGGGTTTGTGTTCACGCTCACCGTGCCCAGTGAGCGAGGACTGCAGCGTAGACGCTTTGTCCAGAATGCCTTAAATGGGAATGGAGACCCAAACAACATGGACAGGGCAGTCAAACTATACAGGAAGCTGAAAAGAGAGATAACATTCCATGGGGCTAAAGAGGTTGCACTCAGCTATTCAACCGGTGCACTTGCCAGTTGCATGGGTCTCATATACAACAGGATGGGAACGGTAACCACAGAAGTAGCTTTTGGCCTGGTGTGTGCCACTTGTGAGCAGATTGCTGACTCACAGCATCGATCTCACAGACAGATGGTGACTACCACCAACCCACTAATCAGGCATGAAAACAGAATGGTGCTGGCCAGCACTACAGCTAAGGCTATGGAGCAGATGGCTGGATCGAGTGAACAGGCAGCGGAAGCCATGGAGGTTGCTAGTCAGGCTAGGCAGATGGTGCAGGCAATGAGGACAATTGGGACTCACCCTAGCTCCAGTGCCGGTCTGAAAGATGATCTTCTTGAAAATTTGCAGGCCTACCAGAAGCGGATGGGAGTGCAAATGCAGCGATTCAAGTGATCCTCTCGTTATTGCCGCAAGTATCATTGGGATCTTGCACTTGATATTGTGGATTCTTGATCGTCCTTTCTTCAAATGTATTTATCGTCGCCTTAAATACGGTTTGAAAAGAGGGCCTTCTACGGAAGGAGTGCCTGAGTCTATGAGGGAAGAGTATCGGCAGGAACAGCAGAGTGCTGTGGATGTTGACGATGGTCATTTTGTCAACATAGAGCTGGAGTAAAAAACTACCTTGTTTCTACT
    ''')

    x = pcr(f,r,t)

    s = parse('''
        >524_pFA6aF (29-mer)
        cacatacgatttaggtgacactatagaac
        >523_AgTEF1tpR (21-mer)
        ggttgtttatgttcggatgtg
        ''')

    from pydna import Dseq,Dseqrecord,read

    t= Dseq("ncacatacgatttaggtgacactatagaaccacatccgaacataaacaacc",
             "gtgtatgctaaatccactgtgatatcttggtgtaggcttgtatttgttgg"[::-1])

    t = Dseqrecord(t)
    t.name="hej"

    #print pcr(s,t,limit=13)

    template = Dseqrecord("tacactcaccgtctatcattatctactatcgactgtatcatctgatagcac")
    print template.features

    p1 =read(">p1\ntacactcaccgtctatcattatc", ds = False)
    p2 =read(">p2\ngtgctatcagatgatacagtcg", ds = False)

    a=Anneal((p1,p2), template)

    print a.products

    from pydna_helper import ape
    from pprint import pprint
    pprint(a.products[0].features)

    #r = Anneal(s, t)

    #print r.template

    #p = r.products[0]

    #print p.figure()

    #print len(p)

    #print p.program()

    #from pydna_helper import ape

    #ape(r.template)

    #print r.products()
#    prod = r.products[0]
#    print type(prod)
#    print prod.detailed_figure()



#    from pydna import read

#    template = Dseqrecord("tacactcaccgtctatcattatctactatcgactgtatcatctgatagcac")
#    p1 = read(">p1\ntacactcaccgtctatcattatc", ds= False)
#    p2 = read(">p1\ncgactgtatcatctgatagcac", ds = False).reverse_complement()
#    amplicon_list = Reaction((p1,p2), template).products
#    amplicon = amplicon_list.pop()
#    print amplicon.detailed_figure()








#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Read version numbers, author etc..
__version__ = "Undefined"
for line in open('pydna/__init__.py'):
    if line.startswith('__'):
        exec(line.strip())

# Change line ending to windows for all text files
import os
for root, dirs, files in os.walk(os.path.abspath(os.path.dirname(__file__))):
    for name in files:
        if not name.lower().endswith(".txt"):
            continue
        filename = os.path.join(root, name)
        with open(filename, "rb") as f:
            data = f.read()                        #.decode("utf-8")
        temp = data.replace('\r\n', '\n')
        temp = temp.replace('\r', '\n')
        temp = temp.replace('\n', '\r\n')
        if not data == temp:
            with open(filename, "wb") as f:
                f.write(temp)
                print("changed", filename)
                

from setuptools import setup #, find_packages

import textwrap, sys

setup(  name='pydna',
        version         =__version__,
        author          =__author__,
        author_email    =__email__,
        #cmdclass = {'build_ext': optional_build_ext},
        #packages = find_packages(),
        packages=['pydna',
                  'pydna.py_rstr_max',],

        #ext_modules=[Extension('pydna.find_sub_strings', ['pydna/findsubstrings_numpy_arrays_cython.c'])],
        #scripts=[],
        url='http://pypi.python.org/pypi/pydna/',
        license='LICENSE.txt',
        description='''Contains classes and code for representing double
                     stranded DNA and functions for simulating homologous
                     recombination between DNA molecules.''',
        long_description=open('README.txt').read(),
        install_requires =[ "networkx>=1.7", "biopython>=1.63","prettytable>=0.7.2"],
        test_suite="run_tests.load_my_tests",
        #include_dirs = [numpy.get_include()],
        zip_safe = False,
        keywords = "bioinformatics",
        classifiers = ['Development Status :: 4 - Beta',
                       'Environment :: Console',
                       'Intended Audience :: Education',
                       'Intended Audience :: Developers',
                       'Intended Audience :: Science/Research',
                       'License :: OSI Approved :: BSD License',
                       'Programming Language :: Python :: 2.7',
                       'Topic :: Education',
                       'Topic :: Scientific/Engineering :: Bio-Informatics',]
        )     
        



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 2013,2014 by Björn Johansson.  All rights reserved.
# This code is part of the Python-dna distribution and governed by its
# license.  Please see the LICENSE.txt file that should have been included
# as part of this package.

'''
    Python-dna
    ~~~~~~~~~~

    The Python-dna package.

    :copyright: Copyright 2013,2014 by Björn Johansson. All rights reserved.
    :license:   This code is part of the Python-dna distribution and governed by its
                license.  Please see the LICENSE.txt file that should have been included
                as part of this package.

'''

__version__      = "0.6.5"
__date__         = "2014-07-31"
__author__       = "Björn Johansson"
__copyright__    = "Copyright 2013,2014 Björn Johansson"
__credits__      = ["Björn Johansson", "Mark Budde"]
__license__      = "BSD"
__maintainer__   = "Björn Johansson"
__email__        = "bjorn_johansson@bio.uminho.pt"
__status__       = "Development" # "Production" #"Prototype"

from pydna.amplify                                  import Anneal
from pydna.amplify                                  import pcr
from pydna.assembly                                 import Assembly
from pydna.download                                 import Genbank
from pydna.dsdna                                    import Dseq
from pydna.dsdna                                    import Dseqrecord
from pydna.dsdna                                    import parse
from pydna.dsdna                                    import parse2
from pydna.dsdna                                    import read
#from pydna.dsdna                                    import glob
from pydna.editor                                   import Editor
from pydna.findsubstrings_suffix_arrays_python      import common_sub_strings
from pydna.primer_design                            import cloning_primers
from pydna.primer_design                            import assembly_primers
from pydna.primer_design                            import primer_pair
from pydna.utils                                    import copy_features
from pydna.utils                                    import eq
from pydna.utils                                    import shift_origin
from pydna.utils                                    import pairwise
from pydna.utils                                    import cseguid


try:
    del findsubstrings_suffix_arrays_python
    del assembly
    del dsdna
    del editor
    del amplify
    del utils
    del download
    del _simple_paths8
    del py_rstr_max
except NameError:
    pass

